{ *************************************************************************************
   Pruebas del giroscopio MPU-6050 conectado a un Arduino Nano y al PC por USB
  *************************************************************************************
   JIDS 07/07/2023
}
unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls,
  synaser;

type
  { THilo }

  THilo = class(TThread)
  private
    fStatusText: string;
    procedure ShowStatus;
  protected
    procedure Execute; override;
  public
    constructor Create(CreateSuspended: boolean);
  end;

  { TForm1 }

  TForm1 = class(TForm)
    Btn_DESCONECTAR: TButton;
    Btn_CONECTAR: TButton;
    Btn_SALIR: TButton;
    Btn_LIMPIAR: TButton;
    E_CONTROLSOFTWARE: TCheckBox;
    E_BITS: TEdit;
    E_BITSPARADA: TComboBox;
    E_CONTROLHARDWARE: TCheckBox;
    E_VELOCIDAD: TComboBox;
    E_PUERTO: TEdit;
    E_PARIDAD: TComboBox;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    L_CONEXION: TLabel;
    Memo1: TMemo;
    PaintBox1: TPaintBox;
    procedure Btn_CONECTARClick(Sender: TObject);
    procedure Btn_DESCONECTARClick(Sender: TObject);
    procedure Btn_LIMPIARClick(Sender: TObject);
    procedure Btn_SALIRClick(Sender: TObject);
    procedure Desconectar;
    procedure FormCreate(Sender: TObject);
  private
  public
    function main : integer;
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

var
  ser : TBlockSerial;
  Hilo1 : THilo;
  linea, lineabytes : string;


  // Para obtener el valor de los campos en una línea tipo CSV o en un TStringList
  function valor(cadena : ansistring; parametro_num : byte; separador_campos : char) : ansistring;
  var
    st : ansistring;
    i : integer;
    contador : byte;
  begin
    st := '';
    contador := 1;
    for i := 1 to Length(cadena) do
    begin
      if cadena[i] = separador_campos then
        contador := contador + 1;
      if contador = parametro_num then
        if cadena[i]<>separador_campos then
          st := st + cadena[i];
    end;
    valor := st;
  end;


procedure TForm1.Btn_CONECTARClick(Sender: TObject);
var
  velocidad : integer;
  puerto  : string;
  paridad : char;
  numbits, bitsparada : byte;
  controlsoft, controlhard : boolean;
begin
  puerto := trim(E_PUERTO.Text);
  velocidad := StrToInt(E_VELOCIDAD.Text);
  numbits := StrToInt(E_BITS.Text);
  paridad := E_PARIDAD.Text[1];
  if E_BITSPARADA.Text = '1'   then bitsparada := SB1;
  if E_BITSPARADA.Text = '1.5' then bitsparada := SB1andHalf;
  if E_BITSPARADA.Text = '2'   then bitsparada := SB2;
  controlsoft := E_CONTROLSOFTWARE.Checked;
  controlhard := E_CONTROLSOFTWARE.Checked;

  ser := TBlockSerial.Create;
  try
    Memo1.Lines.Add('Conectando con Arduino por el puerto USB...');
    ser.LinuxLock := False; // Es para que el archivo de bloque no quede bloqueado si el programa se cierra antes de cerrar el puerto
    ser.Connect(puerto); // COM1: COM2: (Windows) o '/dev/ttyUSB0' (Linux)
    Sleep(25);
    ser.ConvertLineEnd := true;      // Se añade un <CR> al final de cada línea
    // ser.config(115200, 8, 'N', SB1, False, False);
    ser.config(velocidad, numbits, paridad, bitsparada, controlsoft, controlhard);
    sleep(2000);   // No menos de 500ms (para 9600 baudios) 1500 para 115200bd
    L_CONEXION.Caption := 'CONECTADO';
    L_CONEXION.Color := clGreen;
    Btn_CONECTAR.Enabled := false;
    Btn_DESCONECTAR.Enabled := true;
    Memo1.Lines.Add('Conexión establecida con Arduino');
  except
    L_CONEXION.Caption := 'DESCONECTADO';
    L_CONEXION.Color := clGray;
    ShowMessage('[001] Error al abrir el puerto');
  end;

  Hilo1 := THilo.Create(True); // Con True la thread no se inicia automáticamente
  if Assigned(Hilo1.FatalException) then
    raise Hilo1.FatalException;

  // En este punto se inicializan todo lo necesario antes de que las threads se ejecuten
  Hilo1.Start;
  Application.ProcessMessages;
end;

procedure TForm1.Desconectar;
begin
  try
    Hilo1.Terminate;
  except
    ShowMessage('Error al terminar el hilo de ejecución que recibe datos');
  end;

  try
    ser.free;
    L_CONEXION.Caption := 'DESCONECTADO';
    L_CONEXION.Color := clGray;
    Btn_CONECTAR.Enabled := true;
    Btn_DESCONECTAR.Enabled := false;
  except
    ShowMessage('[002] Error al cerrar el puerto');
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  // DecimalSeparator := '.';
end;

procedure TForm1.Btn_DESCONECTARClick(Sender: TObject);
begin
  Desconectar;
end;

procedure TForm1.Btn_LIMPIARClick(Sender: TObject);
begin
  Memo1.Lines.Clear;
end;

procedure TForm1.Btn_SALIRClick(Sender: TObject);
begin
  Application.Terminate;  // No usar Close para finalizar el Thread
end;

{ THilo }

procedure THilo.ShowStatus;
// Este método es llamado por Synchronize(@ShowStatus) y, por lo tanto, ejecutado por el hilo principal
// El thread principal puede acceder a los elementos de los Forms
var
  v1, v2, v1anterior, v2anterior, v1diferencia, v2diferencia : double;
  i : integer;
begin
  v1 := 0.0;
  v2 := 0.0;
  v1anterior := 0.0;
  v2anterior := 0.0;
  v1diferencia := 0.0;
  v2diferencia := 0.0;

  // Form1.Caption := fStatusText;
  // Form1.Memo2.Lines.Add(fStatusText);
  // Application.ProcessMessages;
  //Form1.PaintBox1.Canvas.Refresh;
  //Application.ProcessMessages;

  if length(linea) > 0 then
    if pos('|', linea) > 0 then
    begin
      try
        v1 := StrToFloat(valor(linea, 1, '|'));
      except
        v1 := 0;
      end;
      try
        v2 := StrToFloat(valor(linea, 2, '|'));
      except
        v2 := 0;
      end;

      v1diferencia := v1 - v1anterior;
      v2diferencia := v2 - v2anterior;

      v1anterior := v1 + v1diferencia;
      v2anterior := v2 + v2diferencia;

      Form1.PaintBox1.Canvas.Brush.Style := bsSolid;
      Form1.PaintBox1.Color := clSilver;
      Form1.PaintBox1.Canvas.Clear;

      // Dibuja el color de fondo del gráfico
      Form1.PaintBox1.Canvas.Brush.Color := clSkyBlue;
      Form1.PaintBox1.Canvas.FillRect(0, 0, Form1.PaintBox1.Width, Form1.PaintBox1.Height div 2);
      Form1.PaintBox1.Canvas.Brush.Color := TColor($87B8DE);
      Form1.PaintBox1.Canvas.FillRect(0, Form1.PaintBox1.Height div 2, Form1.PaintBox1.Width, Form1.PaintBox1.Height);
      Form1.PaintBox1.Color := clSilver;

      // Dibuja los ejes de coordenadas
      Form1.PaintBox1.Canvas.Pen.Color := clWhite;
      Form1.PaintBox1.Canvas.Pen.Width := 1;
      Form1.PaintBox1.Canvas.Line( 0,
                                   (Form1.PaintBox1.Height div 2),
                                   Form1.PaintBox1.Width,
                                   (Form1.PaintBox1.Height div 2) );

      Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2),
                                   0,
                                   (Form1.PaintBox1.Width div 2),
                                   Form1.PaintBox1.Height );

      for i := 1 to Form1.PaintBox1.Width do
        if i mod 10 = 0 then
        begin
          Form1.PaintBox1.Canvas.Line( i,
                                       (Form1.PaintBox1.Height div 2) - 5,
                                       i,
                                       (Form1.PaintBox1.Height div 2) + 5 );
        end;

      for i := 1 to Form1.PaintBox1.Height do
        if i mod 10 = 0 then
        begin
          Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2) - 5,
                                       i,
                                       (Form1.PaintBox1.Width div 2) + 5,
                                       i);
        end;
      Form1.PaintBox1.Canvas.Refresh;

      if v1 < 0 then
        begin
          Form1.Memo1.Lines.Add('Bajando morro: ' + FloatToStr(v1diferencia) + ' grados');
          Form1.PaintBox1.Canvas.Pen.Color := clBlue;
          Form1.PaintBox1.Canvas.Pen.Width := 5;
          Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2) - 25,
                                       (Form1.PaintBox1.Height div 2) + abs(round(v1)),
                                       (Form1.PaintBox1.Width div 2) + 25,
                                       (Form1.PaintBox1.Height div 2) + abs(round(v1)) );
          Form1.PaintBox1.Canvas.Refresh;
        end
      else
        begin
          Form1.Memo1.Lines.Add('Subiendo morro: ' + FloatToStr(v1diferencia) + ' grados');
          Form1.PaintBox1.Canvas.Pen.Color := clBlue;
          Form1.PaintBox1.Canvas.Pen.Width := 5;

          Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2) - 25,
                                       (Form1.PaintBox1.Height div 2) - abs(round(v1)),
                                       (Form1.PaintBox1.Width div 2) + 25,
                                       (Form1.PaintBox1.Height div 2) - abs(round(v1)) );
          Form1.PaintBox1.Canvas.Refresh;
        end;

      if v2 < 0 then
        begin
          Form1.Memo1.Lines.Add('Inclinando a izquierda: ' + FloatToStr(v2diferencia) + ' grados');
          Form1.PaintBox1.Canvas.Pen.Color := clRed;
          Form1.PaintBox1.Canvas.Pen.Width := 5;
          Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2) - abs(round(v2)),
                                       (Form1.PaintBox1.Height div 2) - 25,
                                       (Form1.PaintBox1.Width div 2) - abs(round(v2)),
                                       (Form1.PaintBox1.Height div 2) + 25 );
          Form1.PaintBox1.Canvas.Refresh;
        end
      else
        begin
          Form1.Memo1.Lines.Add('Inclinando a derecha: ' + FloatToStr(v2diferencia) + ' grados');
          Form1.PaintBox1.Canvas.Pen.Color := clRed;
          Form1.PaintBox1.Canvas.Pen.Width := 5;
          Form1.PaintBox1.Canvas.Line( (Form1.PaintBox1.Width div 2) + abs(round(v2)),
                                       (Form1.PaintBox1.Height div 2) - 25,
                                       (Form1.PaintBox1.Width div 2) + abs(round(v2)),
                                       (Form1.PaintBox1.Height div 2) + 25 );
          Form1.PaintBox1.Canvas.Refresh;
        end;
    end;

  Application.ProcessMessages;
  linea := '';
  lineabytes := '';
end;

procedure THilo.Execute;
var
  newStatus : string;
begin
  fStatusText := 'Iniciando el hilo que recibe datos del puerto USB ...';
  Synchronize(@Showstatus);
  fStatusText := 'El hilo de recepción de datos del puerto USB está en ejecución ...';
  while (not Terminated) and (true) do
  begin
    try
      Form1.Main;
    except
      on e:Exception do
        ShowMessage(e.Message);
    end;

    // Código del bucle del thread principal
    newStatus:='Hora: '+FormatDateTime('YYYY-MM-DD HH:NN:SS',Now);

    if NewStatus <> fStatusText then
    begin
      fStatusText := newStatus;
      Synchronize(@Showstatus);
    end;
    // sleep(50); // Alternativamente aquí el thread podría esperar por algún evento
    sleep(2); // Alternativamente aquí el thread podría esperar por algún evento
    // Application.ProcessMessages;
  end;
end;

constructor THilo.Create(CreateSuspended: boolean);
begin
  FreeOnTerminate := True;
  inherited Create(CreateSuspended);
end;

function TForm1.main : integer;
var
  s, t : ansistring;
  i : longint;
  cambiar : boolean;
begin
  try
    if (ser.WaitingData <> 0) then
    begin
      s := '';
      s := ser.RecvPacket(5000);  // El parámetro es el TimeOut
      // sleep(10);
      linea := '';
      lineabytes := '';

      if s <> '' then
      begin
        for i := 1 to length(s) do
        begin
          lineabytes := lineabytes + IntToStr(ord(s[i])) + ' ';
        end;
      end;

      if pos(#13#10, s) > 0 then          // Tomamos solo el texto hasta el primer CR-LF
        s := copy(s, 1, pos(#13#10, s)-1);

      s := StringReplace(s, #9, '|', [rfReplaceAll, rfIgnoreCase]);
      s := StringReplace(s, ' ', '', [rfReplaceAll, rfIgnoreCase]);

      // Pongo la Coma como separador de decimales en Windows
      s := StringReplace(s, '.', ',', [rfReplaceAll, rfIgnoreCase]);

      s := trim(s);
      linea := '';

      if s <> '' then
      begin
        for i := 1 to length(s) do
        begin
          if ord(s[i]) > 27 then
            linea := linea + s[i];
        end;
        linea := linea + '|';
      end;
    end;
  except
    linea := '[005] Error al intentar escuchar el puerto';
  end;
end;

end.

